//====================================================================
//  FOUR.CPP
//
//  Four-In-A-Row Game
//
//	This C++ source code for an Excalibur BBS Plug-In is intended only as
//	a sample of using the Alchemist's Workbench API and creating a plug-in
//	for Excalibur.  Excalibur Communications makes no warranty of any kind,
//  expressed or implied, including without limitation, any warranties of
//	merchantability and/or fitness for a particular purpose.  Use this
//	source code at your own risk.
//
//====================================================================

#include <classlib\listimp.h>
#include <owl\owlpch.h>
#include "exlib.h"
#include "four.h"
#include "cmdfour.h"
#include "bwcc.h"

//============================================================================/
// Some local definitions and shortcuts                                       /
//============================================================================/

// these are the playing board dimensions
// and bitmap dimensions
#define GRIDX 8
#define GRIDY 8
#define BMPX 32
#define BMPY 32

#define STRING(id) (string( _hInstance, id ).c_str())
#define FONTH ExDlgFontH(LibH)
#define FONTW ExDlgFontW(LibH)
#define EDITH ExDlgEditH(LibH)

static BOOL Done = FALSE;	// for window cleanup later

//============================================================================/
// Structure for keeping info on who is here to play games with               /
//============================================================================/

struct UserEntry {
	long Id;
	char Handle[41];
	};

BOOL operator==( const UserEntry &be1, const UserEntry &be2 )
	{ return be1.Id == be2.Id; }
typedef TListImp<UserEntry> UserDict;
typedef TListIteratorImp<UserEntry> UserDictImp;

//============================================================================/
// Some DLL overhead                                                          /
//============================================================================/

extern "C" {
	int FAR PASCAL _export LibMain( HINSTANCE, WORD, WORD, LPSTR );
	int FAR PASCAL _export WEP(int);
	WORD FAR PASCAL _export PlugInit( EXLIBH handle, LPCSTR args );
	void FAR PASCAL _export PlugDestroy( EXLIBH handle );
	WORD FAR PASCAL _export PlugCommand( EXLIBH handle, LPBYTE cmd, WORD len );
	WORD FAR PASCAL _export PlugProcess( EXLIBH handle );
	};

extern HINSTANCE _hInstance;
int FAR PASCAL
LibMain( HINSTANCE /* hInstance */, WORD, WORD, LPSTR )
	{
	return TRUE;
	}


int FAR PASCAL
WEP(int /*bSystemExit*/)
	{
	return 1;
	}


//============================================================================/
// TMyEdit is an edit class with special CR behaviour                         /
//============================================================================/

class TMyEdit : public TEdit {
public:
	TMyEdit( TWindow *parent, int id, LPCSTR text, int x, int y, int cx, int cy, int limit )
		: TEdit( parent, id, text, x, y, cx, cy, limit ) {}
	void EvChar( UINT, UINT, UINT );
	DECLARE_RESPONSE_TABLE( TMyEdit );
	};

DEFINE_RESPONSE_TABLE1(TMyEdit,TEdit)
	EV_WM_CHAR,
END_RESPONSE_TABLE;

void TMyEdit::
EvChar( UINT code, UINT, UINT )
	{
	if ( code == 13 ) {
		Parent->SendMessage( WM_COMMAND, CM_SEND_TEXT );
		}
	else
		DefaultProcessing();
	}


//============================================================================/
// The game window class header                                               /
//============================================================================/

class GameWindow : public TWindow {
	EXLIBH		LibH;						// handle to library information
	TMyEdit *	EntryEd;					// for user's response
	TEdit *		CommentEd;				// where the opponent's comments go
	UserDict 	UserList;				// list of who is here
	TListBox *	UserLb;					// show who is here
	TStatic *	PageNote;				// little bit'o information for the user
	TStatic *	OtherNote;				// little bit'o information for the user
	TStatic *	WhoNote;					// little bit'o information for the user
	BOOL			HaveOpponent;			// we playing anyone yet?
	BYTE			Grid[GRIDY][GRIDX];	// current game condition
	long			OppId;					// who we're playing (in case they leave)
	char 			OppName[50];			// convenience for painting the screen
	BOOL			Blue;						// Am I blue? (should I cry?)
	BOOL			MyTurn;					// is it my turn?
public:
	GameWindow( EXLIBH lib );
	~GameWindow()
		{
		Done = TRUE;
		ExSetToDie(LibH);
		ExSetPtr( LibH, 0 );
		OppName[0] = 0;
		}
	void SetupWindow();

	// menu handling
	void CmSendText();
	void CmInvite();
	void CmHelp();
	void EnInGame(  TCommandEnabler &en );
	void EnWaiting( TCommandEnabler &en );

	// other windows responses
	void EvLButtonDown( UINT, TPoint& );
	void Paint( TDC &dc, BOOL clear, TRect &r );

	// routines for processing data from Camelot
	WORD Command( BYTE *cmd, WORD len );
	void cmdTextBlock( BYTE *text );
	void cmdPerson( BYTE *cmd );
	void cmdLeave( BYTE *cmd );
	void cmdError( BYTE *cmd );
	void cmdInvite( BYTE *cmd );
	void cmdColor( BYTE *cmd );
	void cmdMove( BYTE *cmd );

	// support routines
	long getid( char *name );
	LPCSTR getname( long id );
	BOOL checkWinner();
	void doneGame();

	DECLARE_RESPONSE_TABLE( GameWindow );
	};


DEFINE_RESPONSE_TABLE1( GameWindow, TWindow )
	EV_WM_CHAR,
	EV_COMMAND( CM_SEND_TEXT, CmSendText ),
	EV_COMMAND( CM_INVITE, CmInvite ),
	EV_COMMAND( CM_GAMEHELP, CmHelp ),
	EV_COMMAND_ENABLE( CM_INVITE, EnWaiting ),
	EV_COMMAND_ENABLE( CM_SEND_TEXT, EnInGame ),
	EV_WM_LBUTTONDOWN,
END_RESPONSE_TABLE;


GameWindow::GameWindow( EXLIBH lib )
	: LibH(lib), TWindow( 0, "" ), HaveOpponent(0)
	{
	Attr.W = GRIDX*BMPX+220;	// make it look good
	Attr.H = GRIDY*BMPY + 96;

	// initially visible gizmos
	WhoNote = new TStatic( this, -1, STRING(S_PEOPLEHERE),
		FONTW, 6, FONTW*20, FONTH, -1 );
	UserLb = new TListBox( this, 101, FONTW, EDITH+6, 20*FONTW, 200 );
	UserLb->Attr.Style |= LBS_NOINTEGRALHEIGHT;

	// initially INvisible gizmos
	PageNote = new TStatic( this, -1, STRING( S_PAGENOTE ),
		GRIDX*BMPX+32, 86, 200, FONTH, -1 );
	PageNote->Attr.Style &= ~WS_VISIBLE;

	EntryEd = new TMyEdit( this, 102, "",
				GRIDX*BMPX+32, 86+FONTH, 170, EDITH, 200 );
	EntryEd->Attr.Style &= ~WS_VISIBLE;

	OtherNote = new TStatic( this, -1, STRING( S_OPPNOTE ),
		GRIDX*BMPX+32, 86+2*EDITH, 200, FONTH, -1 );
	OtherNote->Attr.Style &= ~WS_VISIBLE;

	CommentEd = new TEdit( this, 103, "",
				GRIDX*BMPX+32, 86+3*EDITH, 170, GRIDY*BMPY+16 - (86+3*EDITH),
				0, TRUE );
	CommentEd->Attr.Style &= ~(WS_VISIBLE|WS_HSCROLL|ES_AUTOHSCROLL);
	CommentEd->Attr.Style |= ES_READONLY;
	}

void _FAR setfont( TWindow *w, void *f )
	{
	w->SendMessage( WM_SETFONT, (WPARAM)(*(HFONT*)f), FALSE );
	}

void GameWindow::SetupWindow()
	{
	TWindow::SetupWindow();

	// force the assorted controls to use the common font
	HFONT font = ExDlgFont( LibH );
	ForEach( setfont, &font );

	UserLb->SetFocus();
	UserLb->SetSelItemRange( TRUE, 0, 0 );
	}


// returns the ID of the user found in the list
long GameWindow::getid( char *name )
	{
	UserDictImp scan( UserList );
	while( scan ) {
		if ( strcmp( scan.Current().Handle, name )==0 ) {
			return scan.Current().Id;
			}
		scan++;
		}
	return -1;
	}

// returns pointer to name given the user's ID
LPCSTR GameWindow::getname( long id )
	{
	UserDictImp scan( UserList );
	while( scan ) {
		if ( scan.Current().Id==id ) {
			return scan.Current().Handle;
			}
		scan++;
		}
	return 0;
	}

// do display cleanup and notifications
void GameWindow::doneGame()
	{
	HaveOpponent=FALSE;
	BYTE code = CMD_GAMEOVER;
	ExAddBlock( LibH, &code, 1 );
	OppId = -1;
	Invalidate();
	WhoNote->ShowWindow( SW_SHOW );
	UserLb->ShowWindow( SW_SHOW );
	PageNote->ShowWindow( SW_HIDE );
	EntryEd->ShowWindow( SW_HIDE );
	OtherNote->ShowWindow( SW_HIDE );
	CommentEd->ShowWindow( SW_HIDE );
	}

// returns TRUE if there are 4-in-a-row anywhere
BOOL GameWindow::checkWinner()
	{
	// check horizontal
	int x, y;
	for ( y=0; y<GRIDY; y++ ) {
		for ( x=0; x<GRIDX-3; x++ ) {
			int which = Grid[y][x];
			int count = 0;
			while( which && Grid[y][x+count]==which && count < 4)
				count++;
			if ( count == 4 )
				return TRUE;
			}
		}
	// check vertical
	for ( x=0; x<GRIDX; x++ ) {
		for ( y=0; y<GRIDY-3; y++ ) {
			int which = Grid[y][x];
			int count = 0;
			while( which && Grid[y+count][x]==which && count < 4)
				count++;
			if ( count == 4 )
				return TRUE;
			}
		}
	// check right-slope
	for( x = 0; x<GRIDX-3; x++ ) {
		for ( y=0; y<GRIDY-3; y++ ) {
			int which = Grid[y][x];
			int count = 0;
			while( which && Grid[y+count][x+count]==which && count < 4)
				count++;
			if ( count == 4 )
				return TRUE;
			}
		}
	// check left-slope
	for( x = 3; x<GRIDX; x++ ) {
		for ( y=0; y<GRIDY-3; y++ ) {
			int which = Grid[y][x];
			int count = 0;
			while( which && Grid[y+count][x-count]==which && count < 4)
				count++;
			if ( count == 4 )
				return TRUE;
			}
		}
	return FALSE;
	}

//============================================================================/
// Assorted menu responses                                                    /
//============================================================================/

// send a comment to the opponent
void GameWindow::CmSendText()
	{
	char text[255];
	EntryEd->GetText( text+1, sizeof( text )-1 );
	EntryEd->SetText( "" );

	if ( ! text[1] ) {
		BWCCMessageBox( HWindow, STRING( S_ENTRYERROR), STRING(S_ERROR),
			MB_OK|MB_ICONSTOP );
		EntryEd->SetFocus();
		return;
		}
	text[0] = CMD_TEXTBLOCK;
	ExAddBlock( LibH, text, strlen( text+1 ) + 2 );
	}

// invite someone to play
void GameWindow::CmInvite()
	{
	int who = UserLb->GetSelIndex();
	if ( who < 0 ) {
		MessageBeep(0);
		return;
		}
	// use OppName when inviting since we'll need it later
	UserLb->GetSelString( OppName, sizeof(OppName));
	struct {
		BYTE cmd;
		long id;
		} data = { CMD_INVITE, getid(OppName) };
	OppId = data.id;		// note it in case they wander off later
	ExAddBlock( LibH, &data, sizeof(data));
   char temp[200];
	wsprintf( temp, STRING( S_DIDINVITE ), OppName );
	BWCCMessageBox( HWindow, temp, STRING( S_NOTICE ), MB_OK|MB_ICONINFORMATION );
	}

void GameWindow::CmHelp()
	{
   WinHelp( "FOUR.HLP", HELP_CONTENTS, 0 ); 
	}

void GameWindow::EnInGame( TCommandEnabler &en )
	{
	en.Enable( HaveOpponent );
	}

void GameWindow::EnWaiting( TCommandEnabler &en )
	{
	en.Enable( ! HaveOpponent );
	}

//============================================================================/
// Routines for processing commands from Camelot                              /
//============================================================================/

WORD GameWindow::Command( BYTE *cmd, WORD )
	{
	switch( *cmd++ ) {
		case CMD_WHOISON: {
			UserEntry &user = *(UserEntry*)cmd;
			UserList.Add( user );
			UserLb->AddString( user.Handle );
			break;
			}
		case CMD_PERSON:
			cmdPerson( cmd );
			break;
		case CMD_LEAVING:
			cmdLeave( cmd );
			break;
		case CMD_ERROR:
			cmdError( cmd );
			break;
		case CMD_INVITE:
			cmdInvite( cmd );
			break;
		case CMD_COLOR:
			cmdColor( cmd );
			break;
		case CMD_MOVE:
			cmdMove( cmd );
			break;
		case CMD_TEXTBLOCK:
			cmdTextBlock( cmd );
			break;
		}
	return AL_OK;
	}

// add this person to the list of people here
void GameWindow::cmdPerson( BYTE *cmd )
	{
	struct tag {
		long id;
		char handle[1];
		} &person = *(tag*)cmd;

	UserEntry user;
	user.Id = person.id;
	strcpy( user.Handle, person.handle );
	UserLb->AddString( user.Handle );
	UserList.Add( user );
	}

// the person with this ID has left
void GameWindow::cmdLeave( BYTE *cmd )
	{
	UserDictImp scan( UserList );
	while( scan ) {
		if ( scan.Current().Id == *(long*)cmd ) {
			int index = UserLb->FindExactString( scan.Current().Handle, 0 );
			UserLb->DeleteString( index );
			UserList.Detach( scan.Current() );
			break;
			}
		scan++;
		}
	if ( OppId == *(long*)cmd ) {		// hey! that was my opponent
		MessageBeep( MB_ICONEXCLAMATION );
		BWCCMessageBox( HWindow, STRING( S_DEFAULT ), STRING( S_NOTICE ),
						MB_OK|MB_ICONEXCLAMATION );
		doneGame();
		}
	}

// error message
void GameWindow::cmdError( BYTE *cmd )
	{
	MessageBeep( MB_ICONEXCLAMATION );
	BWCCMessageBox( HWindow, STRING( ERROR_BASE+*cmd ), STRING( S_ERROR ), MB_OK|MB_ICONEXCLAMATION );
	}

// you've been invited
void GameWindow::cmdInvite( BYTE *cmd )
	{
	LPCSTR who;
	who = getname( *(long*)cmd );
	if ( ! who ) {							// odd?  why are we getting this?
		BYTE cmd = CMD_REJECT;			// Some kind of mistake?
		ExAddBlock( LibH, &cmd, 1 );	// do an automatic reject
		return;
		}

	BYTE code = CMD_REJECT;
	char temp[200];
	wsprintf( temp, STRING(S_INVITE), who );
	MessageBeep( MB_ICONQUESTION );
	if ( BWCCMessageBox( HWindow, temp, STRING( S_NOTICE ), MB_OKCANCEL|MB_ICONQUESTION ) == IDOK ) {
		code = CMD_ACCEPT;
		strcpy( OppName, who );
		OppId = *(long*)cmd;	// note it now in case they leave later
		}
	ExAddBlock( LibH,  &code, 1 );
	}

// Notification of your piece color.
// This starts all games.
void GameWindow::cmdColor( BYTE *cmd )
	{
	Blue = *cmd;
	MyTurn = !*cmd;
	HaveOpponent = TRUE;
	memset( Grid, 0, sizeof( Grid ) );

	WhoNote->ShowWindow( SW_HIDE );
	UserLb->ShowWindow( SW_HIDE );
	PageNote->ShowWindow( SW_SHOW );
	EntryEd->ShowWindow( SW_SHOW );
	EntryEd->SetFocus();
	OtherNote->ShowWindow( SW_SHOW );
	CommentEd->ShowWindow( SW_SHOW );
	Invalidate();
	}

// got a move from your opponent
void GameWindow::cmdMove( BYTE *cmd )
	{
	Grid[cmd[1]][cmd[0]] = Blue ? 2 : 1;
	MyTurn = TRUE;
	Invalidate();
	if ( checkWinner() ) {	// had to have been the other player
		MessageBeep( MB_ICONEXCLAMATION );
		BWCCMessageBox( HWindow, STRING( S_LOSE ), STRING( S_SORRY ), MB_OK|MB_ICONEXCLAMATION );
		doneGame();
		}
	}

// got comment from opponent
void GameWindow::cmdTextBlock( BYTE *text )
	{
	CommentEd->SetSelection( 30000, 30000 );	// go to the end effectively
	CommentEd->Insert( (char*)text );
   CommentEd->Insert( "\r\n" );
	}

//============================================================================/
// Windows interface routines                                                 /
//============================================================================/

void GameWindow::Paint( TDC &dc, BOOL clear, TRect &r )
	{
	TWindow::Paint( dc, clear, r );
	if ( ! HaveOpponent )
		return;
	// draw the board
	dc.SelectStockObject( LTGRAY_BRUSH );
	dc.SelectStockObject( NULL_PEN );
	dc.Rectangle( 8, 8, GRIDX*BMPX+27, GRIDY*BMPY+27 );
	dc.SelectStockObject( BLACK_PEN );
	// draw the game grid
	for ( int i=1; i<GRIDX+1; i++ ) {
		dc.MoveTo( i*BMPX, 16 );
		dc.LineTo( i*BMPX, GRIDY*BMPY+17 );
		}
	for ( i=1; i<GRIDY+1; i++ ) {
		dc.MoveTo( 16, i*BMPY );
		dc.LineTo( GRIDX*BMPX+17, i*BMPY );
		}
	// now give it a nice border
	dc.SelectObject( TPen( TColor::Gray ) );
	dc.MoveTo( 15, GRIDY*BMPY+16 );
	dc.LineTo( 15, 15 );
	dc.LineTo( GRIDX*BMPX+17, 15 );
	dc.MoveTo( GRIDX*BMPX+25, 9 );
	dc.LineTo( GRIDX*BMPX+25, GRIDY*BMPY+25 );
	dc.LineTo( 8, GRIDY*BMPY+25 );
	dc.SelectStockObject( WHITE_PEN );
	dc.MoveTo( 8, GRIDY*BMPY+24 );
	dc.LineTo( 8, 8 );
	dc.LineTo( GRIDX*BMPX+25, 8 );
	dc.MoveTo( 16, GRIDY*BMPY+17 );
	dc.LineTo( GRIDX*BMPX+17, GRIDY*BMPY+17 );
	dc.LineTo( GRIDX*BMPX+17, 15 );

	// draw the pieces
	TBitmap blue( _hInstance, 1 );
	TBitmap red( _hInstance, 2 );
	TMemoryDC memdc( dc );
	for ( int y=0; y<GRIDY; y++ ) {
		for ( int x=0; x<GRIDX; x++ ) {
			TRect dest( x*BMPX+16, y*BMPY+16, x*BMPX+48, y*BMPY+48 );
			if ( Grid[y][x]==1 ) {
				memdc.SelectObject( blue );
				dc.BitBlt( dest, memdc, TPoint(0,0) );
				}
			else if ( Grid[y][x]==2 ) {
				memdc.SelectObject( red );
				dc.BitBlt( dest, memdc, TPoint(0,0) );
				}
			}
		}
	// draw names and things
	dc.SelectStockObject( LTGRAY_BRUSH );
	dc.SelectStockObject( NULL_PEN );
	dc.Rectangle( GRIDX*BMPX+32, 16, GRIDX*BMPX+202, 82 );
	TBitmap plainred( _hInstance, IDB_PLAINRED );
	TBitmap plainblue( _hInstance, IDB_PLAINBLUE );
	TBitmap go( _hInstance, IDB_GO );
	TRect dest( GRIDX*BMPX+48, 16, GRIDX*BMPY+80, 48 );
	if ( Blue )
		memdc.SelectObject( plainblue );
	else
		memdc.SelectObject( plainred );
	dc.BitBlt( dest, memdc, TPoint(0,0));
	dest.Offset( 0, BMPY );
	if ( Blue )
		memdc.SelectObject( plainred );
	else
		memdc.SelectObject( plainblue );
	dc.BitBlt( dest, memdc, TPoint(0,0));

	dest.Set( GRIDX*BMPX+33, 16, GRIDX*BMPX+49, BMPY+16 );
	if ( ! MyTurn )
		dest.Offset( 0, BMPY );
	memdc.SelectObject( go );
   dc.BitBlt( dest, memdc, TPoint(0,0));

	dc.SelectObject( TPen( TColor::Gray ) );
	dc.MoveTo( GRIDX*BMPX+33, 80 );
	dc.LineTo( GRIDX*BMPX+200, 80 );
	dc.LineTo( GRIDX*BMPX+200, 16 );

	dc.SelectStockObject( WHITE_PEN );
	dc.MoveTo( GRIDX*BMPX+32, 79 );
	dc.LineTo( GRIDX*BMPX+32, 16 );
	dc.LineTo( GRIDX*BMPX+200, 16 );

	dc.SetBkMode( TRANSPARENT );
	dest.Set( GRIDX*BMPX+80, 16, GRIDX*BMPX+200, BMPY+16 );
	dc.DrawText( STRING(S_YOU), -1, dest, DT_SINGLELINE|DT_VCENTER );
	dest.Offset( 0, BMPY );
	dc.DrawText( OppName, -1, dest, DT_SINGLELINE|DT_VCENTER );
	}

void GameWindow::
EvLButtonDown( UINT, TPoint &pt )
	{
	if ( !MyTurn ) {
		MessageBeep(0);
		return;
		}
	if ( pt.x < 16 || pt.x > GRIDX*BMPX+16
		|| pt.y < 16 || pt.y > GRIDY*BMPY+16 ) {
		MessageBeep(0);
		return;
		}
	int x=(pt.x-16)/BMPX;
	int y=(pt.y-16)/BMPY;
	if ( Grid[y][x] ) {
		MessageBeep(0);
		return;
		}
	Grid[y][x] = Blue ? 1 : 2;
	MyTurn = FALSE;
	Invalidate();
	char temp[3];
	temp[0] = CMD_MOVE;
	temp[1] = x;
	temp[2] = y;
	ExAddBlock( LibH, temp, 3 );

	if ( checkWinner() ) {	// it can only be this player
		MessageBeep( MB_ICONEXCLAMATION );
		BWCCMessageBox( HWindow, STRING( S_YOUWIN ), STRING( S_HEY ),
						MB_OK|MB_ICONEXCLAMATION );
		doneGame();
		}
	}


//============================================================================/
// The actual Excalibur interface                                             /
//============================================================================/

void _export SetupTools( EXLIBH handle, EXCHILD vchild )
	{
	EXTOOLBAR cb = ExCreateToolbar( handle, vchild, -1 );
	ExButtonGadget( cb, MAKEINTRESOURCE(IDB_INVITE), CM_INVITE, EXBG_COMMAND,
						FALSE, EXBG_STATE_UP, FALSE );
	ExButtonGadget( cb, MAKEINTRESOURCE(IDB_ENTER), CM_SEND_TEXT, EXBG_COMMAND,
						FALSE, EXBG_STATE_UP, FALSE );
	ExButtonGadget( cb, MAKEINTRESOURCE(IDB_HELP), CM_GAMEHELP, EXBG_COMMAND,
						FALSE, EXBG_STATE_UP, FALSE );
	ExInsertToolbar( vchild, cb );
	}


WORD FAR PASCAL _export PlugInit( EXLIBH handle, LPCSTR /* args */ )
	{
	ExcalWindow win;
	win.DLLInstance = _hInstance;
	win.OwlWindow = new GameWindow( handle );
	win.Title = "Four-In-A-Row";
	win.Class = "OwlWindow";
	win.Menu = MAKEINTRESOURCE( 1 );
	win.Icon = MAKEINTRESOURCE( 1 );
	win.KBHandling = FALSE;
	win.ShowMethod = SW_SHOW;
	win.ToolbarSetup = SetupTools;
	ExCreateTool( handle, &win );
	ExSetPtr( handle, win.OwlWindow );
	return AL_OK;
	}


void FAR PASCAL _export PlugDestroy( EXLIBH handle )
	{
	GameWindow *win = (GameWindow*)ExGetPtr( handle );
	if ( win )
		win->CloseWindow();
	while ( ! Done )
		ExPumpMessages( handle );
	}


#pragma argsused
WORD FAR PASCAL _export PlugCommand( EXLIBH handle, LPBYTE cmd, WORD len )
	{
	GameWindow *win = (GameWindow*)ExGetPtr( handle );
	return win->Command( cmd, len );
	}

WORD FAR PASCAL _export PlugProcess( EXLIBH /* handle */ )
	{
	return AL_OK;
	}


